
require "util"
local task = require("tasks")
local walk_proximity = 0.1
local debug_state = true

local state = 1
local state_reached = 0
local idle = 0

local player
local player_selection
local position
local destination
local target_inventory
local walking
local step
local count
local item
local direction
local input
local output
local type

local drop_item
local drop_position

local function debug(msg)
	if debug_state then
		player.print(msg)
	end
end

-- Check that the entity can be selected and is within reach
local function check_selection_reach()
	player.update_selected_entity(position)
	player_selection = player.selected

	if not player_selection then
		if not walking.walking then
			debug(string.format("Step %d - %s: Cannot select entity", state, step))
		end

		return false
	end

	if not player.can_reach_entity(player_selection) then
		if not walking.walking then
			debug(string.format("Step %d - %s: Cannot reach entity", state, step))
		end

		return false
	end

	return true
end

-- Check that it is possible to get the inventory of the entity
local function check_inventory() 
	target_inventory = player_selection.get_inventory(slot)

	if not target_inventory then
		if not walking.walking then
			debug(string.format("Step: %d - %s: Cannot get entity inventory", state, step))
		end

		return false
	end

	return true
end

-- Place an item from the character's inventory into an entity's inventory
-- Returns false on failure to prevent advancing state until within reach
-- It is possible to put 0 items if none are in the character's inventory
local function put()

	if not check_selection_reach() then
		return false;
	end

	if not check_inventory() then
		return false;
	end

	if amount == -1 then 
		amount = player.get_item_count(item)
	else 
		amount = math.min(player.get_item_count(item), amount)
	end

    if amount == 0 then
		debug(string.format("Step %d - Put: Nothing to put", state))
		return true
	end

	amount = target_inventory.insert{name=item, count=amount}

	if amount == 0 then
		debug(string.format("Step %d - Put: Entity is already full", state))
		return true
	end

	player.remove_item{name=item, count=amount}
	return true
end

-- Creating buildings
local function build()

	if player.get_item_count(item) == 0 then
		if(state > state_reached) then 
			debug(string.format("Step %d - Build: %s not available", state, item:gsub("-", " "):gsub("^%l", string.upper)))
			state_reached = state
		end

		return false
	end

	if player.can_place_entity{name = item, position = position} then
		if player.surface.can_fast_replace{name = item, position = position, direction = direction, force = "player"} then
			if player.surface.create_entity{name = item, position = position, direction = direction, force="player", fast_replace=true, player=player} then
				state = state - 1
				player.remove_item({name = item, count = 1})
				return true
			end
		else
			if player.surface.create_entity{name = item, position = position, direction = direction, force="player"} then
				player.remove_item({name = item, count = 1})
				return true
			end
		end

	else 
		if not walking.walking then
			debug(string.format("Step %d - Build: %s cannot be placed", state, item:gsub("-", " "):gsub("^%l", string.upper)))
		end

		return false
	end
end

local function walk(delta_x, delta_y)
	if delta_x > walk_proximity then
		-- Easterly
		if delta_y > walk_proximity then
			return {walking = true, direction = defines.direction.southeast}
		elseif delta_y < -walk_proximity then
			return {walking = true, direction = defines.direction.northeast}
		else
			return {walking = true, direction = defines.direction.east}
		end
	elseif delta_x < -walk_proximity then
		-- Westerly
		if delta_y > walk_proximity then
			return {walking = true, direction = defines.direction.southwest}
		elseif delta_y < -walk_proximity then
			return {walking = true, direction = defines.direction.northwest}
		else
			return {walking = true, direction = defines.direction.west}
		end
	else
		-- Vertically
		if delta_y > walk_proximity then
			return {walking = true, direction = defines.direction.south}
		elseif delta_y < -walk_proximity then
			return {walking = true, direction = defines.direction.north}
		else
			return {walking = false, direction = defines.direction.north}
		end
	end
end

-- Routing function to perform one of the many available tasks
-- True: Indicates the calling function should advance the state. 
-- False: Indicates the calling function should not advance state.
local function doTask(task)
	if task[1] == "build" then
		step = "Build"
		position = task[2]
		item = task[3]
		direction = task[4]

		return build()

	elseif task[1] == "put" then
		step = "Put"
		position = task[2]
		item = task[3]
		amount = task[4]
		slot = task[5]

		return put(task[2], task[3], task[4], task[5])

	end
end

-- Main per-tick event handler
script.on_event(defines.events.on_tick, function(event)

    if not player then 
		player = game.players[1]
		position = player.position
		destination = {x = position.x, y = position.y}
		player.force.research_queue_enabled = true
	end

	position = player.position

	item = "stone-furnace"

	if player.get_item_count(item) < 2 then
		player.insert({name=item, count=30})
	end

	item = "coal"

	if player.get_item_count(item) < 2 then
		player.insert({name=item, count=30})
	end

	-- primary movement
	walking = walk(destination.x - position.x, destination.y - position.y)
	if walking.walking == false then

		if task[state][1] == "walk" then
			destination = {x = task[state][2][1], y = task[state][2][2]}
			walking = walk(destination.x - position.x, destination.y - position.y)
			state = state + 1

		elseif task[state][1] == "mine" then
			player.update_selected_entity(task[state][2])
			player.mining_state = {mining = true, position = task[state][2]}

		elseif doTask(task[state]) then
			-- Do task while standing still
			state = state + 1

		end
	else
		if task[state][1] ~= "walk" and task[state][1] ~= "mine" and task[state][1] ~= "idle" and task[state][1] ~= "pick" then
			if doTask(task[state]) then
				-- Do task while walking
				state = state + 1
			end
		end
	end

	player.walking_state = walking
end)

script.on_event(defines.events.on_player_mined_entity, function(event)

	state = state + 1
end)

-- Skips the freeplay intro
script.on_event(defines.events.on_game_created_from_scenario, function()

	remote.call("freeplay", "set_skip_intro", true)

end)